"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import argparse
import copy
from . import const
from .chromosome import ChromosomeData, ChromosomeDict
from .io import get_dtype_length
import logging
import numpy as np  # type: ignore
import os
import pickle
import pandas as pd  # type: ignore
from typing import Any, Dict, List


class Record:
    """DataBase Record"""

    _data: Dict[str, Any]

    def __init__(self, record: bytes, column_dtypes: Dict[str, str]):
        super(Record, self).__init__()
        self.__parse_bytes(record, column_dtypes)

    @property
    def data(self) -> Dict[str, Any]:
        return copy.copy(self._data)

    def __parse_bytes(self, record: bytes, column_dtypes: Dict[str, str]) -> None:
        """Parse record bytes.

        Parses a record from its bytes based on column dtypes.

        Arguments:
            record {bytes} -- record bytes
            column_dtypes {Dict[str, str]} -- column dtypes
        """
        if len(record) != get_dtype_length(column_dtypes):
            raise AssertionError
        self._data = {}
        current_location = 0
        for label in const.database_columns:
            dtype = column_dtypes[label]
            byte_size = get_dtype_length({label: dtype})
            self._data[label] = np.frombuffer(
                record[current_location : (current_location + byte_size)], dtype
            )[0]
            current_location += byte_size

    def to_dataframe(self) -> pd.DataFrame:
        return pd.DataFrame(self._data, index=[0])

    def to_csv(self, sep: str = ",") -> str:
        csv_fields: List[str] = []
        for x in const.database_columns:
            data = self._data[x]
            if isinstance(data, bytes):
                csv_fields.append(str(data.decode()))
            else:
                csv_fields.append(str(data))
        return sep.join(csv_fields)

    @staticmethod
    def __norm_value_in_range(v, r):
        if r[1] == 0:
            return np.nan
        return (v - r[0]) / (r[1] - r[0])

    def __update_score_by_nOT(self, F):
        off_target_no = self._data["off_target_no"]
        if off_target_no <= F[0]:
            self._data["score"] = 0
            return None
        if off_target_no > F[1]:
            self._data["score"] = np.inf
            return None
        return self.__norm_value_in_range(off_target_no, F)

    def __update_score_by_dG_Tm(self, Gs):
        ss_dG = self._data["ss_dG"]
        tm_dG = self._data["Tm_dG"]
        if ss_dG >= tm_dG * min(Gs):
            self._data["score"] = 0
            return None
        if ss_dG < tm_dG * max(Gs):
            self._data["score"] = np.inf
            return None
        return self.__norm_value_in_range(ss_dG, [tm_dG * f for f in Gs])

    def __update_score_by_dG_Gs(self, Gs):
        ss_dG = self._data["ss_dG"]
        if ss_dG >= Gs[0]:
            self._data["score"] = 0
            return None
        if ss_dG < Gs[1]:
            self._data["score"] = np.inf
            return None
        return self.__norm_value_in_range(ss_dG, Gs)

    def add_score(self, F, Gs) -> None:
        norm_nOT = self.__update_score_by_nOT(F)
        if norm_nOT is None:
            return None
        if all(x >= 0 for x in Gs):
            norm_ss_dG = self.__update_score_by_dG_Tm(Gs)
        else:
            norm_ss_dG = self.__update_score_by_dG_Gs(Gs)
        if norm_ss_dG is None:
            return None
        self._data["score"] = np.mean([norm_nOT, norm_ss_dG])
        return None

    def __getitem__(self, key: str) -> Any:
        allowed_columns = ["score", *const.database_columns]
        if key not in allowed_columns:
            raise KeyError(f"unrecognized key '{key}'")
        return self.data[key]

    def __repr__(self) -> str:
        return str(self._data)


class DataBase:
    """Buffering and checking class for ifpd2 database."""

    _root: str
    _args: Dict[str, Any]
    _chromosomes: ChromosomeDict
    _dtype: Dict[str, str]
    _record_byte_size: int

    def __init__(self, path: str):
        """
        Arguments:
            path {str} -- absolute path to database folder
        """
        super(DataBase, self).__init__()
        if not os.path.isdir(path):
            raise AssertionError(f"cannot find database folder '{path}'")
        db_pickle_path = os.path.join(path, "db.pickle")
        if not os.path.isfile(db_pickle_path):
            raise AssertionError(f"'db.pickle is missing in '{path}'")

        with open(db_pickle_path, "rb") as IH:
            details = pickle.load(IH)
            self._chromosomes = details["chromosomes"]
            self._dtype = details["dtype"]
            if isinstance(details["args"], argparse.Namespace):
                self._args = self.__parse_old_pickle(details["args"])
            else:
                self._args = dict(details["args"])

        if not isinstance(self._chromosomes, ChromosomeDict):
            raise AssertionError
        for chromosome in self.chromosome_list:
            chromosome_path = os.path.join(path, f"{chromosome.decode()}.bin")
            if not os.path.isfile(chromosome_path):
                raise AssertionError(
                    f"missing expected chromosome file: '{chromosome_path}'"
                )

        self._record_byte_size = get_dtype_length(self._dtype)
        if self._record_byte_size <= 0:
            raise AssertionError

        self._root = path

    @property
    def path(self):
        return self._root

    @property
    def chromosome_list(self) -> List[bytes]:
        return self._chromosomes.keys()

    @property
    def chromosome_sizes_nt(self) -> Dict[bytes, int]:
        return self._chromosomes.sizes_nt

    # @property
    # def chromosome_sizes_bytes(self) -> Dict[bytes, int]:
    #     return self._chromosomes.sizes_bytes

    @property
    def chromosome_recordnos(self) -> Dict[bytes, int]:
        return self._chromosomes.recordnos

    @property
    def record_byte_size(self) -> int:
        return self._record_byte_size

    @property
    def dtype(self) -> Dict[str, str]:
        return copy.copy(self._dtype)

    @staticmethod
    def __parse_old_pickle(args: argparse.Namespace) -> Dict[str, Any]:
        _args = vars(args)
        return dict(
            output_path=_args["output"],
            off_target_paths=_args["hush"],
            melting_temperature_paths=_args["melting"],
            secondary_structure_paths=_args["secondary"],
            bin_size=_args["binsize"],
            prefix=_args["prefix"],
        )

    def get_chromosome(self, chromosome: bytes) -> ChromosomeData:
        return self._chromosomes.get_chromosome(chromosome)

    def log_details(self) -> None:
        """Log database details."""
        logging.info(f"Database name: {self._args['output_path']}")
        logging.info(f"Sequence max length: {self._dtype['sequence'][2:]}")
        logging.info("")
        logging.info("[bold]## Input files[/bold]")
        logging.info(f"hush files: {self._args['off_target_paths']}")
        logging.info(f"oligo-melting files: {self._args['melting_temperature_paths']}")
        logging.info(f"OligoArrayAux files: {self._args['secondary_structure_paths']}")
        logging.info("")
        logging.info("[bold]## Chromosome details[/bold]")
        logging.info(f"Expecting {len(self._chromosomes)} chromosomes.")
        logging.info("Chromosomes:")
        for chromosome, data in self._chromosomes.items():
            empty_label = "".join(" " for c in chromosome.decode())
            logging.info(f"\t{chromosome.decode()} => size : {data.size_nt} (nt)")
            logging.info(f"\t{chromosome.decode()} => size : {data.size_bytes} (bytes)")
            logging.info(f"\t{empty_label} => recordno : {data.recordno}")
        logging.info("")
        logging.info("[bold]Record details[/bold]")
        logging.info(f"Record size in bytes: {self._record_byte_size}")
        logging.info(f"Dtype: {self._dtype}")



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from dataclasses import dataclass
from os.path import isdir, isfile
from typing import Optional, Tuple


def path_exists(path: str) -> bool:
    return isdir(path) or isfile(path)


@dataclass(frozen=True)
class Folder:
    path: str
    exists: bool = False

    def __post_init__(self):
        if (
            self.exists
            and not isdir(self.path)
            or not self.exists
            and path_exists(self.path)
        ):
            raise AssertionError((self.path, self.exists))


@dataclass(frozen=True)
class File:
    path: str
    exists: bool = False

    def __post_init__(self):
        if (
            self.exists
            and not isfile(self.path)
            or not self.exists
            and path_exists(self.path)
        ):
            raise AssertionError


@dataclass(frozen=True)
class PositiveInteger:
    n: int

    def __post_init__(self):
        if self.n < 1:
            raise AssertionError


@dataclass(frozen=True)
class NonNegativeFloat:
    n: float

    def __post_init__(self):
        if self.n < 0:
            raise AssertionError


@dataclass(frozen=True)
class NonNegativeInteger:
    n: int

    def __post_init__(self):
        if self.n <= 0:
            raise AssertionError


@dataclass(frozen=True)
class PositiveFloat:
    n: float
    limit: Optional[float] = None
    limit_included: bool = True

    def __post_init__(self):
        if self.limit is None:
            if self.n <= 0:
                raise AssertionError

        elif self.limit_included:
            if not 0 < self.n <= self.limit:
                raise AssertionError
        elif not 0 < self.n < self.limit:
            raise AssertionError


@dataclass(frozen=True)
class GenomicRegion:
    start: int
    end: int

    def __post_init__(self):
        if self.start < 0:
            raise AssertionError
        if self.end < self.start and self.end != -1:
            raise AssertionError

    def astuple(self) -> Tuple[int, int]:
        return (self.start, self.end)


@dataclass(frozen=True)
class NonNegativeIntInterval:
    start: int
    end: int

    def __post_init__(self):
        if self.start < 0:
            raise AssertionError
        if self.end < self.start:
            raise AssertionError

    def astuple(self) -> Tuple[int, int]:
        return (self.start, self.end)


@dataclass(frozen=True)
class QueryWindow:
    size: Optional[int]
    shift: Optional[float]

    def __post_init__(self):
        if self.size is not None and self.size < 1:
            raise AssertionError
        if not 0 < self.shift <= 1:
            raise AssertionError

    def astuple(self) -> Tuple[Optional[int], Optional[float]]:
        return (self.size, self.shift)


@dataclass(frozen=True)
class QueryFocus:
    size: float
    step: float

    def __post_init__(self):
        if self.size <= 0:
            raise AssertionError
        if self.step <= 0:
            raise AssertionError


@dataclass(frozen=True)
class FreeEnergyInterval:
    start: float
    end: float

    def astuple(self) -> Tuple[float, float]:
        return (self.start, self.end)


@dataclass(frozen=True)
class GCRange:
    low: float
    high: float

    def __post_init__(self):
        if not 0 <= self.low <= 1:
            raise AssertionError
        if not 0 <= self.high <= 1:
            raise AssertionError
        if self.low > self.high:
            raise AssertionError



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import logging
from rich.logging import RichHandler  # type: ignore
from rich.console import Console  # type: ignore
import os
from typing import Optional


def add_log_file_handler(path: str, logger_name: Optional[str] = None) -> None:
    if os.path.isdir(path):
        raise AssertionError
    log_dir = os.path.dirname(path)
    if not (os.path.isdir(log_dir) or log_dir == ""):
        raise AssertionError
    fh = RichHandler(console=Console(file=open(path, mode="w+")), markup=True)
    fh.setLevel(logging.INFO)
    logging.getLogger(logger_name).addHandler(fh)
    logging.getLogger(logger_name).info(f"[green]Log to[/]: '{path}'")



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import click  # type: ignore
import os
from rich import print  # type: ignore
import shlex
from shutil import copyfile
import sys

from .. import __path__, __version__
from ..const import CONTEXT_SETTINGS


SHELL_TYPES = ("bash", "zsh", "fish")


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    "--shell-type",
    "-s",
    help="shell type for which to activate autocompletion",
    type=click.Choice(SHELL_TYPES, case_sensitive=False),
    default="bash",
    show_default=True,
)
@click.option(
    "--regenerate",
    help="to regenerate autocompletion file, mainly for developers",
    type=click.BOOL,
    default=False,
    is_flag=True,
)
@click.version_option(__version__)
def main(shell_type: str, regenerate: bool) -> None:
    user_home_path = os.path.expanduser("~")
    autocomplete_path = os.path.join(
        __path__[0], "autocomplete", f".ifpd2-complete.{shell_type}"
    )

    if regenerate:
        regenerate_autocompletion_files(shell_type, autocomplete_path)

    if shell_type in {"bash", "zsh"}:
        autocomplete_bash_or_zsh(user_home_path, autocomplete_path, shell_type)
    elif shell_type == "fish":
        autocomplete_fish(user_home_path, autocomplete_path)

    print("Done. :thumbs_up: :smiley:")


def regenerate_autocompletion_files(shell_type: str, autocomplete_path: str) -> None:
    if shell_type not in SHELL_TYPES:
        raise AssertionError(f"unrecognized shell type '{shell_type}'.")
    cmd = f"_IFPD2_COMPLETE={shell_type}_source ifpd2 > {autocomplete_path}"
    os.system(shlex.quote(cmd))
    print(f"Regenerated {shell_type} completion file: {autocomplete_path}")


def autocomplete_fish(user_home_path: str, autocomplete_path: str) -> None:
    destination_path = os.path.join(
        user_home_path, ".config/fish/completions/ifpd2.fish"
    )

    if os.path.isfile(destination_path):
        print("Autocompletion was previously set up. Skipping.")
        sys.exit()

    copyfile(
        autocomplete_path,
        destination_path,
    )


def autocomplete_bash_or_zsh(
    user_home_path: str, autocomplete_path: str, shell_type: str = "bash"
) -> None:
    if shell_type not in {"bash", "zsh"}:
        raise AssertionError

    autocompletion_string = f". {autocomplete_path} # IFPD2-AUTOCOMPLETE\n"
    run_command_path = os.path.join(user_home_path, f".{shell_type}rc")

    with open(run_command_path, "r") as OH:
        if autocompletion_string in OH.readlines():
            print("Autocompletion was previously set up. Skipping.")
            sys.exit()

    with open(run_command_path, "a+") as OH:
        OH.write(f"\n{autocompletion_string}")



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import configparser as cp
import numpy as np  # type: ignore
import os
import pandas as pd  # type: ignore
import shutil

from . import asserts as ass


class OligoProbe:
    """Converts a DataFrame of oligo data into an OligoProbe."""

    def __init__(self, oligo_data):
        super(OligoProbe, self).__init__()
        self.data = oligo_data

    @property
    def data(self):
        return self._data.copy()

    @data.setter
    def data(self, oligo_data):
        if not isinstance(oligo_data, pd.DataFrame):
            raise AssertionError
        required_columns = ["start", "end", "Tm"]
        if any(col not in oligo_data.columns for col in required_columns):
            raise AssertionError
        self._data = oligo_data
        self._range = (self._data["start"].min(), self._data["end"].max())
        self._size = self._range[1] - self._range[0]
        oDists = self._data["start"].values[1:] - self._data["end"].values[:-1]
        self._spread = np.std(oDists)
        self._d_range = (oDists.min(), oDists.max())
        self._d_mean = oDists.mean()
        self._tm_range = (self._data["Tm"].min(), self._data["Tm"].max())

    @property
    def n_oligos(self):
        return self._data.shape[0]

    @property
    def path(self):
        return self._data.index

    @property
    def range(self):
        return self._range

    @property
    def tm_range(self):
        return self._tm_range

    @property
    def d_range(self):
        return self._d_range

    @property
    def d_mean(self):
        return self._d_mean

    @property
    def spread(self):
        return self._spread

    @property
    def size(self):
        return self._size

    @property
    def featDF(self):
        df = pd.DataFrame(
            [
                self.range[0],
                self.range[1],
                self.n_oligos,
                self.size,
                self.spread,
                self.d_range[0],
                self.d_range[1],
                self.d_mean,
                np.diff(self.tm_range)[0],
            ]
        ).transpose()
        df.columns = [
            "start",
            "end",
            "nOligos",
            "size",
            "spread",
            "d_min",
            "d_max",
            "d_mean",
            "tm_range",
        ]
        return df

    def __repr__(self):
        rep = f"<OligoProbe[{self.range[0]}:{self.range[1]}"
        rep += f":{self.size}:{self.spread}]>"
        return rep

    def count_shared_oligos(self, probe):
        # Counts oligos shared with another probe
        # based on their paths
        return np.intersect1d(self.path, probe.path).shape[0]

    def export(self, path):
        if os.path.isfile(path):
            raise AssertionError
        if os.path.isdir(path):
            shutil.rmtree(path)
        os.mkdir(path)

        self.featDF.to_csv(os.path.join(path, "probe.tsv"), "\t", index=False)
        self.data.to_csv(os.path.join(path, "oligos.tsv"), "\t", index=False)

        with open(os.path.join(path, "probe.fa"), "w+") as BH:
            for i in range(self.data.shape[0]):
                oligo = self.data.iloc[i]
                BH.write(
                    "".join(
                        [
                            f">{oligo['name']}:",
                            f"{oligo['chromosome']}:{oligo['start']}-{oligo['end']}\n",
                        ]
                    )
                )
                BH.write(f"{oligo['sequence']}\n")


class OligoPathBuilder:
    """OligoPathBuilder contains methods to select non-overlapping sets of
    oligos (i.e., probes) from an oligo DataFrame in input.
    """

    N = int(48)  # Number of oligos per probe
    D: int = int(2)  # Minimum distance between consecutive oligos
    Tr = 10.0  # Melting temperature range half-width
    Ps = int(10000)  # Probe size threshold, in nt (Ps > 1)
    Ph = 0.1  # Maximum hole size in probe as fraction of probe size
    Po = 0.5  # Probe oligo intersection threshold for path reduction

    def _assert(self):
        ass.ert_type(self.N, int, "N")
        ass.ert_nonNeg(self.N, "N")

        ass.ert_type(self.D, int, "D")
        ass.ert_nonNeg(self.D, "D")

        ass.ert_type(self.Tr, float, "Tr")
        ass.ert_nonNeg(self.Tr, "Tr")

        ass.ert_type(self.Ps, int, "Ps")
        if self.Ps <= 1:
            raise AssertionError

        ass.ert_type(self.Ph, float, "Ph")
        ass.ert_inInterv(self.Ph, 0, 1, "Ph")

        ass.ert_type(self.Po, float, "Po")
        ass.ert_inInterv(self.Po, 0, 1, "Po")

    @staticmethod
    def get_paths(A):
        path_set = set()
        path_start_set = set()

        idxs = [
            i
            for i in range(A.shape[0] - 1)
            if i in path_start_set or len(A[i, i + 1].nonzero()[0]) != 0
        ]

        for i in idxs:
            path = [i]
            k = i
            if 1 not in A[k, (k + 1) :]:
                continue

            j = A[k, (k + 1) :].argmax() + k + 1
            path.append(j)

            while (j + 1) < A.shape[0]:
                k = j
                if 1 not in A[k, (k + 1) :]:
                    break

                j = A[k, (k + 1) :].argmax() + k + 1
                path.append(j)

            path_set.add(tuple(path))
            path_start_set.add(path[0])

        return path_set

    @staticmethod
    def get_non_overlapping_paths(oData, D):
        # Gets all paths of consecutive non-overlapping oligos with minimum
        # distance equal to D
        ass.ert_type(oData, pd.DataFrame, "oData")

        start_positions = oData["start"].values
        end_positions = oData["end"].values

        edges = [
            np.logical_or(
                end_positions + D < oData["start"].values[i],
                start_positions - D >= oData["end"].values[i],
            )
            for i in range(oData.shape[0])
        ]
        return OligoPathBuilder.get_paths(np.vstack(edges).astype("i"))

    def __size_paths(self, path_set, exit_polls):
        sized_paths = set()
        for path in set(path_set):
            if self.N > len(path):
                exit_polls["N"] = exit_polls["N"] + 1
                continue
            if self.N == len(path):
                sized_paths.add(path)
            else:
                for j in range(len(path) - self.N + 1):
                    subpath = path[j : (j + self.N)]
                    sized_paths.add(subpath)
        return sized_paths

    def filter_paths(self, path_set, oData):
        # Selects oligo paths based on length, melting temperature, size, and
        # presence of gaps.
        ass.ert_type(oData, pd.DataFrame, "oData")

        exit_polls = {"P": 0, "N": 0, "S": 0, "H": 0, "T": 0}

        selected_paths = set()
        for path in list(self.__size_paths(path_set, exit_polls)):
            if path in selected_paths:
                continue
            passed, comment = self.__path_passes(path, oData)
            exit_polls[comment] = exit_polls[comment] + 1
            if not passed:
                continue
            selected_paths.add(path)

        comment = "".join(f"{r}{c}" for (c, r) in exit_polls.items())
        return (list(selected_paths), comment)

    def __path_passes(self, path, oData):
        if not isinstance(path, list):
            path = list(path)
        pData = oData.iloc[path, :]
        probe_size = pData["end"].values[-1]
        probe_size -= pData["start"].values[0]
        if self.Ps < probe_size:
            return (False, "S")
        dData = pData["start"].values[1:] - pData["end"].values[:-1]
        max_hole_size = dData.max()
        if self.Ph * probe_size < max_hole_size:
            return (False, "H")
        Tm_range = pData["Tm"].max() - pData["Tm"].min()
        if 2 * self.Tr < Tm_range:
            return (False, "T")
        return (True, "P")

    @staticmethod
    def convert_paths_to_probes(path_list, oData):
        # Converts a list of paths into a list of probes
        ass.ert_type(path_list, list, "path list")
        probe_list = []
        if len(path_list) != 0:
            for path in path_list:
                probe_list.append(OligoProbeBuilder.path2probe(path, oData))
        return probe_list

    @staticmethod
    def path2probe(path, oData):
        # Convert an oligo path into an OligoProbe
        return OligoProbe(oData.iloc[list(path), :])


class OligoProbeBuilder(OligoPathBuilder):
    """Class to build OligoProbe objects."""

    def __init__(self):
        super(OligoProbeBuilder, self).__init__()
        self.k = None
        self.F = [0, 99]  # Threshold on number of off-targets (range)
        self.Gs = [0.0, 0.5]  # dG of SS either as kcal/mol (negative)
        #  or as fraction dG of hybrid (0<=Gs<=1)
        #  (range)
        self.Ot = 0.1  # Step for oligo score relaxation

    @property
    def config(self):
        config = cp.ConfigParser()
        config["AIM"] = {"Oligo(s) number": self.N}
        if not isinstance(self.k, type(None)):
            config["AIM"]["Oligo length (nt)"] = str(self.k)
        config["OLIGO FILTERS"] = {
            "Off-target threshold": self.F,
            "Secondary structure dG threshold": self.Gs,
            "Oligo score relaxation step": self.Ot,
        }
        config["PROBE FILTERS"] = {
            "Melting temperature range (degC)": self.Tr,
            "Min. consecutive oligo distance (nt)": self.D,
            "Probe size threshold": self.Ps,
            "Maximum hole size": self.Ph,
        }
        return config

    def _assert(self):
        OligoPathBuilder._assert(self)

        if not isinstance(self.k, type(None)):
            ass.ert_type(self.k, int, "k")
            ass.ert_nonNeg(self.k, "k")
            if (self.k + self.D) * self.N > self.Ps:
                raise AssertionError

        if isinstance(self.F, tuple):
            self.F = list(self.F)
        ass.ert_type(self.F, list, "F")
        if len(self.F) != 2:
            raise AssertionError
        for i in range(2):
            ass.ert_type(self.F[i], int, f"F[{i}]")
            if self.F[i] < 0:
                raise AssertionError
        if self.F[1] < self.F[0]:
            raise AssertionError

        if isinstance(self.Gs, tuple):
            self.Gs = list(self.Gs)
        ass.ert_type(self.Gs, list, "Gs")
        if len(self.Gs) != 2:
            raise AssertionError
        for i in range(2):
            ass.ert_type(self.Gs[i], float, f"Gs[{i}]")
            if self.Gs[i] > 1:
                raise AssertionError
        if not all(np.array(self.Gs) < 0) and not all(np.array(self.Gs) >= 0):
            raise AssertionError
        if self.Gs[0] >= 0:
            if self.Gs[1] < self.Gs[0]:
                raise AssertionError
        elif self.Gs[1] > self.Gs[0]:
            raise AssertionError

        ass.ert_type(self.Ot, float, "Ot")
        if self.Ot <= 0 or self.Ot > 1:
            raise AssertionError

    def get_prologue(self):
        s = "* OligoProbeBuilder *\n\n"
        s += f"Aim to build probes with {self.N} oligos each.\n"
        s += f"Off-target threshold range set at {self.F}.\n"
        s += "Threshold on the delta free energy of the most stable"
        if isinstance(self.Gs[0], int):
            s += f" secondary structure set at range {self.Gs} kcal/mol.\n"
        else:
            s += " secondary structure\nset at range"
            s += f" {[t*100 for t in self.Gs]}% of the delta free energy of"
            s += " hybridization.\n"

        s += f"\nMelting temperature range of {2*self.Tr} degC.\n"
        s += "Minimum distance between consecutive oligos in a probe"
        s += f" set at {self.D} nt.\n"
        s += f"Probe size threshold set at {self.Ps} nt.\n"
        s += "Reducing probes when oligo intersection fraction is equal to"
        s += f" or greater than {self.Po}.\n"

        return s

    def start(self, oGroup, window, cfr_step, logger):
        self._assert()
        return self.__build_probe_candidates(oGroup, window, cfr_step, logger)

    def __focus_oligos(self, window, oGroup):
        if np.isnan(window["cfr_start"]):
            oGroup.focus_all()
        else:
            oGroup.set_focus_window(window["cfr_start"], window["cfr_end"])
            oGroup.expand_focus_to_n_oligos(self.N)
        return oGroup

    def __build_probe_candidates(self, oGroup, window, cfr_step, logger):
        # Applies oligo filters to the oligo group,
        # expands the focus group if needed, and build probe candidates
        oGroup = self.__focus_oligos(window, oGroup)
        probe_list = self.__explore_filter(oGroup, logger)
        if not np.isnan(window["cfr_start"]):
            while len(probe_list) == 0:
                oGroup.reset_threshold()
                if oGroup.focus_window_size >= self.Ps:
                    oGroup.discard_focused_oligos_safeN(self.N - 1, self.D)
                if not oGroup.expand_focus_by_step(cfr_step):
                    break
                probe_list = self.__explore_filter(oGroup, logger)

        return probe_list

    def __build_probe_list(self, oGroup, noligos, score_thr, max_score, logger):
        nOligos_in_focus_window = oGroup.get_n_focused_oligos(True)

        probe_list = self.__get_non_overlapping_probes(
            oGroup.get_focused_oligos(True), logger
        )
        while len(probe_list) == 0:
            score_thr += self.Ot
            if score_thr > max_score:
                break

            oGroup.apply_threshold(score_thr)
            nOligosUsable = oGroup.get_n_focused_oligos(True)
            if nOligosUsable in [noligos, 0]:
                continue

            logger.info(
                " ".join(
                    [
                        f"Relaxed oligo score threshold to {score_thr:.3f}",
                        f"({nOligosUsable} oligos usable)...",
                    ]
                )
            )
            probe_list = self.__get_non_overlapping_probes(
                oGroup.get_focused_oligos(True), logger
            )

            if nOligosUsable == nOligos_in_focus_window:
                logger.warning("All oligos included. Score relaxation ineffective.")
                break
            noligos = nOligosUsable

        return probe_list

    def __explore_filter(self, oGroup, logger):
        # Explores the 0-to-max_score score threshold range and stops as soon as
        # one probe candidate passes all user-defined thresholds

        if oGroup.focus_window_size < self.Ps:
            max_score = 0
            logger.warning(
                " ".join(
                    [
                        "Score relaxation deactivated when focus region",
                        "size is smaller than probe size threshold.",
                    ]
                )
            )
        else:
            max_score = 1

        score_thr = 0
        oGroup.apply_threshold(score_thr)
        nOligos_prev_score_thr = oGroup.get_n_focused_oligos(True)
        logger.info(
            " ".join(
                [
                    f"Set oligo score threshold at {score_thr:.3f}",
                    f"({nOligos_prev_score_thr} oligos usable)...",
                ]
            )
        )

        if max_score == 0 and nOligos_prev_score_thr == 0:
            return []

        while nOligos_prev_score_thr == 0 and score_thr <= max_score - self.Ot:
            score_thr += self.Ot
            oGroup.apply_threshold(score_thr)
            nOligos_prev_score_thr = oGroup.get_n_focused_oligos(True)
            logger.info(
                " ".join(
                    [
                        f"Relaxed oligo score threshold to {score_thr:.3f}",
                        f"({nOligos_prev_score_thr} oligos usable)...",
                    ]
                )
            )

        return self.__build_probe_list(
            oGroup, nOligos_prev_score_thr, score_thr, max_score, logger
        )

    def __get_non_overlapping_probes(self, oData, logger, verbosity=True):
        # Builds non overlapping oligo paths from an oligo data table, filters
        # them based on class attributes, and then converts the remaining ones
        # to OligoProbe objects.

        paths = self.get_non_overlapping_paths(oData, self.D)
        if len(paths) == 0:
            logger.warning("No oligo paths found.")
            return []

        pathMaxLen = np.max([len(p) for p in list(paths)])
        if verbosity:
            nPaths = len(paths)
            logger.info(
                " ".join(
                    [
                        f"Found {nPaths} sets with up to {pathMaxLen}",
                        "non-overlapping oligos.",
                    ]
                )
            )
        if pathMaxLen < self.N:
            logger.warning("Longest path is shorter than requested. Skipped.")
            return []

        paths, comment = self.filter_paths(paths, oData)
        if verbosity:
            logger.info(
                " ".join(
                    [
                        f"{len(paths)}/{nPaths} oligo paths remaining",
                        f"after filtering. ({comment})",
                    ]
                )
            )

        return self.convert_paths_to_probes(paths, oData)

    def __init_probe_selection(self, sorted_probes):
        selected_probes = []
        probe_ref = sorted_probes[0]
        for probe in sorted_probes[1:-1]:
            n_shared_oligos = probe_ref.count_shared_oligos(probe)

            if probe_ref.n_oligos == n_shared_oligos:
                raise Exception("Encountered probe duplicates!")

            if self.Po * self.N <= n_shared_oligos:
                probe_ref = self.select_probe_from_pair(probe_ref, probe)
            else:
                selected_probes.append(probe_ref)
                probe_ref = probe

        if probe_ref not in selected_probes:
            selected_probes.append(probe_ref)

        return selected_probes, probe_ref

    def reduce_probe_list(self, probe_list):
        try:
            if len(probe_list) == 0:
                return []
            sorted_probes = sorted(probe_list, key=lambda p: p.range[0])

            selected_probes, probe_ref = self.__init_probe_selection(sorted_probes)

            n_shared_oligos = probe_ref.count_shared_oligos(sorted_probes[-1])
            if self.Po * self.N > n_shared_oligos:
                selected_probes.append(sorted_probes[-1])

            return selected_probes
        except Exception as e:
            print(e)
            raise

    @staticmethod
    def select_probe_from_pair(probeA, probeB):
        if probeA.size < probeB.size:
            return probeA
        if np.diff(probeA.tm_range)[0] < np.diff(probeB.tm_range)[0]:
            return probeA
        if probeA.spread / probeA.d_mean < probeB.spread / probeB.d_mean:
            return probeA
        return probeB

    @staticmethod
    def import_probes(ipath):
        # Imports output from given directory path
        if not os.path.isdir(ipath):
            raise AssertionError

        pPath = os.path.join(ipath, "probe_paths.tsv")
        if not os.path.isfile(pPath):
            return []

        oPath = os.path.join(ipath, "oligos.tsv")
        if not os.path.isfile(oPath):
            return []

        oligos = pd.read_csv(oPath, "\t", index_col=0)
        paths = pd.read_csv(pPath, "\t", index_col=0)

        return [
            OligoProbe(oligos.loc[[int(o) for o in p.split(",")]])
            for p in paths.iloc[:, 0]
        ]

    @staticmethod
    def export_probes(probe_list, opath):
        # Exports a list of probes to a directory
        if not os.path.isdir(opath):
            raise AssertionError

        probe_df = pd.concat([p.featDF for p in probe_list], ignore_index=True)
        probe_df.sort_values("start", inplace=True)
        probe_df.to_csv(os.path.join(opath, "probe_feat.tsv"), "\t")

        pd.concat([p.data for p in probe_list]).drop_duplicates().to_csv(
            os.path.join(opath, "oligos.tsv"), "\t"
        )

        probe_paths = [
            [",".join(str(x) for x in probe_list[pi].data.index.tolist())]
            for pi in range(len(probe_list))
        ]

        probe_paths = pd.DataFrame(probe_paths)
        probe_paths.columns = ["cs_oligos"]
        probe_paths.to_csv(os.path.join(opath, "probe_paths.tsv"), "\t")



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import click  # type: ignore
from ...const import CONTEXT_SETTINGS
from ...database import DataBase
from ...walker2 import ChromosomeWalker
import logging
from tqdm import tqdm  # type: ignore


@click.command(
    name="check",
    context_settings=CONTEXT_SETTINGS,
    help="""Check integrity of INPUT database.""",
)
@click.argument("input_paths", metavar="INPUT", nargs=1, type=click.Path(exists=True))
def main(input_paths: str) -> None:
    DB = DataBase(input_paths)
    for chromosome in DB.chromosome_list:
        walker = ChromosomeWalker(DB, chromosome)
        previous_position = -1
        for record in tqdm(
            walker.buffer(),
            desc=f"Checking sorting '{chromosome.decode()}'",
            total=DB.chromosome_recordnos[chromosome],
        ):
            if record["start"] <= previous_position:
                raise AssertionError
            previous_position = record["start"]
    logging.info("That's all! :smiley:")



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from .database import DataBase, Record
from .region import GenomicRegion
import itertools
import logging
import os
from typing import IO, Iterator, List, Union


class ChromosomeWalker:
    """docstring for Walker"""

    __chromosome: bytes
    __db: DataBase
    __IH: IO

    def __init__(self, db: Union[DataBase, str], chromosome: bytes):
        super(ChromosomeWalker, self).__init__()
        self.__db = DataBase(db) if isinstance(db, str) else db
        if chromosome not in self.__db._chromosomes.keys():
            raise AssertionError
        self.__IH = open(
            os.path.join(self.__db._root, f"{chromosome.decode()}.bin"), "rb"
        )
        self.__chromosome = chromosome

    @property
    def db(self) -> DataBase:
        return self.__db

    def __jump_to_bin(self, start_from_nt: int = 0) -> None:
        """Move buffer pointer to the first record of a bin.

        Keyword Arguments:
            start_from_nt {int} -- position in nucleotides (default: {0})
        """
        position_in_bytes = self.__db._chromosomes.get_chromosome(
            self.__chromosome
        ).index[start_from_nt]
        self.__IH.seek(position_in_bytes)

    def read_next_record(self) -> bytes:
        """Read the next record.

        The buffer pointer moves to the start of the following record.

        Returns:
            bytes -- record bytes
        """
        return self.__IH.read(self.__db.record_byte_size)

    # def read_previous_record(self) -> bytes:
    #     """Read the previous record.

    #     The buffer pointer does not move.

    #     Returns:
    #         bytes -- record bytes
    #     """
    #     self.rewind()
    #     return self.read_next_record()

    def read_next_record_and_rewind(self) -> bytes:
        """Reads the next record.

        The buffer pointer does not move.

        Returns:
            bytes -- record bytes
        """
        record = self.read_next_record()
        self.rewind()
        return record

    def next_record_exists(self) -> bool:
        """Tries reading the next record.

        Returns:
            bool -- whether next record exists
        """
        return len(self.read_next_record_and_rewind()) != 0

    def rewind(self) -> None:
        """Rewind of one record.

        The buffer pointer moves to the beginning of the previous record.
        """
        self.__IH.seek(max(self.__IH.tell() - self.__db.record_byte_size, 0))

    # def skip(self) -> None:
    #     """Skip one record.

    #     The buffer pointer moves to the beginning of the following record.
    #     """
    #     self.__IH.seek(self.__IH.tell() + self.__db.record_byte_size)

    def fastforward(self, start_from_nt: int) -> None:
        """Jump to the first record at a given position.

        First the buffer pointer is moved to the beginning of the bin containing the
        start_from_nt position. Then, records are read an parsed until their start is
        greater than the start_from_nt value. Finally, the buffer pointer is moved to
        the beginning of the last parsed record.

        Arguments:
            IH {IO} -- database input handle
            chromosome {bytes} -- chromosome label
            start_from_nt {int} -- position to fastforward to (in nucleotides)
        """
        if start_from_nt == 0:
            self.__IH.seek(0)
            return
        self.__jump_to_bin(start_from_nt)

        record_start = 0
        while record_start < start_from_nt:
            if self.next_record_exists():
                record_start = Record(self.read_next_record(), self.__db.dtype)["start"]
            else:
                logging.warning(
                    " ".join(
                        [
                            "the specified location",
                            f"({self.__chromosome.decode()}:{start_from_nt})",
                            "is outside the database.",
                        ]
                    )
                )
                return

    def buffer(self, start_from_nt: int = 0, end_at_nt: int = -1) -> Iterator[Record]:
        """Buffer a chromosome's records.

        Buffer records from a chromosome within the specified region.
        To buffer the whole records, specify a [0, -1] region.

        Keyword Arguments:
            start_from_nt {int} -- region starting position (default: {0})
            end_at_nt {int} -- region end position (default: {-1})

        Yields:
            Iterator[Record] -- parsed record
        """
        self.fastforward(start_from_nt)
        record = self.read_next_record()
        while len(record) != 0:
            parsed_record = Record(record, self.__db.dtype)
            if parsed_record["start"] > end_at_nt and end_at_nt > 0:
                break
            yield parsed_record
            record = self.read_next_record()

    def walk_single_region(self, region: GenomicRegion) -> Iterator[List[Record]]:
        if region.chromosome != self.__chromosome:
            raise AssertionError
        focus_start, focus_end = region.focus
        record_list = list(self.buffer(focus_start, focus_end))
        yield record_list
        while region.can_increase_focus():
            region.increase_focus()
            new_focus_start, new_focus_end = region.focus
            record_list = list(
                itertools.chain(
                    *[
                        list(self.buffer(new_focus_start, focus_start)),
                        record_list,
                        list(self.buffer(focus_end, new_focus_end)),
                    ]
                )
            )
            yield record_list
            focus_start, focus_end = region.focus

    def walk_multiple_regions(self, region_set_list: List[List[GenomicRegion]]) -> None:
        for region_set_idx in range(len(region_set_list)):
            region_set = region_set_list[region_set_idx]
            for region_idx in range(len(region_set)):
                region = region_set[region_idx]
                for record_set in self.walk_single_region(region):
                    logging.info(
                        " ".join(
                            [
                                f"Window {region_set_idx}.{region_idx}",
                                f"({region.focus}):",
                                f"read {len(record_set)} records",
                            ]
                        )
                    )



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])

dtype_melting = {
    "name": "|S",
    "Tm_dG": "<f4",
    "Tm_dH": "<f4",
    "Tm_dS": "<f4",
    "Tm": "<f4",
    "sequence": "|S",
}
dtype_hush = {"sequence": "|S", "off_target_no": "<u8"}
dtype_secondary = {"ss_dG": "<f4"}
dtype_sequence_features = {"sequence": "|S", "gc_content": "<f4"}
dtype_header_features = {"name": "|S", "chromosome": "|S", "start": "<u4", "end": "<u4"}

DEFAULT_DATABASE_INDEX_BIN_SIZE = 100000

database_columns = [
    "name",
    "chromosome",
    "start",
    "end",
    "sequence",
    "gc_content",
    "off_target_no",
    "Tm_dG",
    "Tm_dH",
    "Tm_dS",
    "Tm",
    "ss_dG",
]



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from copy import copy
from ...const import DEFAULT_DATABASE_INDEX_BIN_SIZE
import logging
from os.path import isdir, isfile
from typing import Any, Dict, List, Set


class DBMakeSettings:
    _output_path: str
    _off_target_paths: Set[str]
    _melting_temperature_paths: Set[str]
    _secondary_structure_paths: Set[str]
    _bin_size: int = DEFAULT_DATABASE_INDEX_BIN_SIZE
    prefix: str = ""

    def __init__(self, output_path: str):
        super(DBMakeSettings, self).__init__()
        self.output_path = output_path
        self._off_target_paths = set()
        self._melting_temperature_paths = set()
        self._secondary_structure_paths = set()

    @property
    def output_path(self) -> str:
        return self._output_path

    @output_path.setter
    def output_path(self, output_path: str) -> None:
        if isdir(output_path) or isfile(output_path):
            raise AssertionError(f"'{output_path}' already exists")
        self._output_path = output_path

    @property
    def bin_size(self) -> int:
        return self._bin_size

    @bin_size.setter
    def bin_size(self, bin_size: int) -> None:
        if bin_size < 1:
            raise AssertionError
        self._bin_size = bin_size

    @property
    def off_target_paths(self) -> Set[str]:
        return copy(self._off_target_paths)

    @property
    def melting_temperature_paths(self) -> Set[str]:
        return copy(self._melting_temperature_paths)

    @property
    def secondary_structure_paths(self) -> Set[str]:
        return copy(self._secondary_structure_paths)

    def add_off_target_path(self, path: str) -> None:
        if not isfile(path):
            raise AssertionError(f"'{path}' not found.")
        self._off_target_paths.add(path)

    def add_off_target_path_list(self, path_list: List[str]) -> None:
        for path in path_list:
            self.add_off_target_path(path)

    def add_melting_temperature_path(self, path: str) -> None:
        if not isfile(path):
            raise AssertionError(f"'{path}' not found.")
        self._melting_temperature_paths.add(path)

    def add_melting_temperature_path_list(self, path_list: List[str]) -> None:
        for path in path_list:
            self.add_melting_temperature_path(path)

    def add_secondary_structure_path(self, path: str) -> None:
        if not isfile(path):
            raise AssertionError(f"'{path}' not found.")
        self._secondary_structure_paths.add(path)

    def add_secondary_structure_path_list(self, path_list: List[str]) -> None:
        for path in path_list:
            self.add_secondary_structure_path(path)

    def init_login(self) -> None:
        logging.info(f"output folder: '{self.output_path}'")
        logging.info(f"chromosome prefix: '{self.prefix}'")
        logging.info(f"hush: {self.off_target_paths}")
        logging.info(f"oligo-melting: {self.melting_temperature_paths}")
        logging.info(f"OligoArrayAux: {self.secondary_structure_paths}")

    def asdict(self) -> Dict[str, Any]:
        return dict(
            output_path=self.output_path,
            off_target_paths=self.off_target_paths,
            melting_temperature_paths=self.melting_temperature_paths,
            secondary_structure_paths=self.secondary_structure_paths,
            bin_size=self.bin_size,
            prefix=self.prefix,
        )



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import copy
from . import asserts as ass
from . import const
import gzip
import logging
import os
import pandas as pd  # type: ignore
from typing import Dict, Iterator, List, Set, Tuple
from tqdm import tqdm  # type: ignore


def get_dtype_length(dtype: Dict[str, str]) -> int:
    """Calculate dtype record length in bytes.

    Arguments:
        dtype {Dict[str, str]} -- column dtype Dict[column_name: column_dtype]

    Returns:
        int -- length of record in bytes
    """
    return sum(int(label.strip("><|SUuif")) for label in dtype.values())


def iterate_lines(path: str) -> Iterator:
    """Iterate lines of a file, even if gzipped."""
    if not path.endswith(".gz"):
        yield from open(path)
    else:
        with gzip.open(path) as IH:
            for line in IH:
                yield line.decode()


def parse_hush(path: str) -> pd.DataFrame:
    """Parse HUSH output.

    Arguments:
        path {str} -- path to HUSH output file

    Returns:
        pd.DataFrame -- parsed HUSH data
    """
    if not os.path.isfile(path):
        raise AssertionError(f"cannot find file '{path}'")
    logging.info(f"parsing: '{path}'")
    sequence_lengths: Set[int] = set()
    parsed_lines: List[Tuple[str, str, int]] = []

    header = ""
    for line in tqdm(iterate_lines(path), desc="Parsing hush output", leave=False):
        if line.startswith(">"):
            header = line.strip()
        else:
            line_split = line.strip().split(",")
            parsed_lines.append((header[1:], line_split[0], int(line_split[1].strip())))
            sequence_lengths.add(len(line_split[0]))
            header = ""

    hush_df = pd.DataFrame(parsed_lines, columns=["name", "sequence", "off_target_no"])
    hush_df.set_index("name", inplace=True)
    ass.ert_in_dtype(hush_df["off_target_no"].values.max(), "u8")
    dtype = copy.copy(const.dtype_hush)
    dtype["sequence"] = f"|S{max(sequence_lengths)}"
    return hush_df.astype(dtype)


def parse_melting(path: str, sep: str = "\t", header: bool = True) -> pd.DataFrame:
    """Parse oligo-melting output.

    Arguments:
        path {str} -- path to oligo-melting output

    Keyword Arguments:
        sep {str} -- column separator (default: {"t"})
        header {bool} -- whether to expect a header line (default: {True})

    Returns:
        pd.DataFrame -- parsed oligo-melting data
    """
    if not os.path.isfile(path):
        raise AssertionError(f"cannot find file '{path}'")
    logging.info(f"parsing: '{path}'")
    expected_columns = copy.copy(const.dtype_melting)
    melting_df = pd.read_csv(path, sep=sep, header=None, skiprows=1 if header else 0)
    if melting_df.shape[1] != len(expected_columns):
        raise AssertionError
    melting_df.columns = list(expected_columns.keys())
    melting_df.set_index("name", inplace=True)
    expected_columns.pop("name", None)
    expected_columns[
        "sequence"
    ] = f'|S{max({len(seq) for seq in melting_df["sequence"].values})}'
    return melting_df.astype(expected_columns)


def parse_secondary(path: str) -> pd.DataFrame:
    """Parse OligoArrayAux .ct output.

    Arguments:
        path {str} -- path to OligoArrayAux .ct output

    Returns:
        pd.DataFrame -- parsed OligoArrayAux .ct data
    """
    if not os.path.isfile(path):
        raise AssertionError(f"cannot find file '{path}'")
    logging.info(f"parsing: '{path}'")
    parsed_lines: List[Tuple[str, float]] = []
    with open(path, "r") as IH:
        for line in tqdm(IH, desc="Parsing OligoArrayAux output", leave=False):
            if "dG = " not in line:
                continue
            _, dG_string, name = line.strip().split("\t")
            parsed_lines.append((name, float(dG_string[5:])))
    secondary_df = pd.DataFrame(parsed_lines, columns=["name", "ss_dG"])
    secondary_df.set_index("name", inplace=True)
    ass.ert_in_dtype(secondary_df["ss_dG"].values.max(), "f4")
    return secondary_df.astype(const.dtype_secondary)


def write_oligos(
    output_path: str,
    oligo_generator: Iterator[Tuple[str, str]],
    desc: str = "writing",
) -> None:
    with open(output_path, "w+") as OH:
        for oligo in tqdm(oligo_generator, desc=desc, leave=False):
            OH.write(">%s\n%s\n" % oligo)



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

import click  # type: ignore
import copy
from ...const import CONTEXT_SETTINGS, DEFAULT_DATABASE_INDEX_BIN_SIZE
from ...const import dtype_header_features, dtype_sequence_features
from ...const import database_columns, dtype_hush, dtype_melting, dtype_secondary
from ... import asserts as ass
from ...chromosome import ChromosomeDict
from ...io import parse_hush, parse_melting, parse_secondary
from .settings import DBMakeSettings
import logging
import numpy as np  # type: ignore
import os
import pandas as pd  # type: ignore
import pickle
from rich.progress import Progress, track  # type: ignore
from typing import Callable, Dict, List, Set, Tuple


@click.command(
    name="make",
    context_settings=CONTEXT_SETTINGS,
    help="""
Assembles files into a database. At least one of the following is required,
to retrieve sequence: hush output (-O), or oligo-melting output (-T).

The name of each record should be in the format: 'name pos=chrom:start-end'.

Accepted formats:
hush            single-line sequence fasta file, where a ", XXX" is appended after the
                  sequence. XXX is the number of detected off-targets.
oligo-melting   .tsv file with six (6) columns: name, dG, dH, dS, Tm, Seq.
                  NOTE: the first line of the file is skipped during parsing because
                  it is expected to be a header line.
OligoArrayAux   .ct (Connectivity Table) file. Details on format available here:
                  https://rna.urmc.rochester.edu/Text/File_Formats.html

Files are merged by: hush header, oligo-melting name, and
OligoArrayAux name (3rd field of 1st column).
""",
)
@click.argument("output", nargs=1, type=click.Path(exists=False))
@click.option(
    "--hush",
    "-O",
    metavar="hush_path",
    nargs=1,
    multiple=True,
    type=click.Path(exists=True),
    help="Path to hush output. Format details above.",
)
@click.option(
    "--melting",
    "-T",
    metavar="hush_oligomelting_pathpath",
    nargs=1,
    multiple=True,
    type=click.Path(exists=True),
    help="Path to oligo-melting output. Format details above.",
)
@click.option(
    "--secondary",
    "-S",
    metavar="oligoarrayaux_path",
    nargs=1,
    multiple=True,
    type=click.Path(exists=True),
    help="Path to OligoArrayAux output. Format details above.",
)
@click.option(
    "--prefix",
    "-p",
    type=click.STRING,
    default="",
    help="Prefix to be added to chromosome labels.",
)
@click.option(
    "--binsize",
    "-b",
    default=DEFAULT_DATABASE_INDEX_BIN_SIZE,
    type=click.INT,
    help=f"Index bin size. Default: {DEFAULT_DATABASE_INDEX_BIN_SIZE}",
)
def main(
    output: str,
    hush: List[str],
    melting: List[str],
    secondary: List[str],
    prefix: str,
    binsize: int,
) -> None:
    settings = DBMakeSettings(output)
    settings.add_off_target_path_list(hush)
    settings.add_melting_temperature_path_list(melting)
    settings.add_secondary_structure_path_list(secondary)
    settings.bin_size = binsize
    settings.prefix = prefix

    if (
        len(settings.off_target_paths) == 0
        and len(settings.melting_temperature_paths) != 0
    ):
        raise AssertionError("please provide either --hush or --melting")

    os.mkdir(settings.output_path)
    dbdf = pd.DataFrame(columns=["name"])
    dbdf.set_index("name", inplace=True)

    dbdf = populate_db(dbdf, settings.off_target_paths, parse_hush, "hush")
    dbdf = populate_db(
        dbdf, settings.melting_temperature_paths, parse_melting, "oligo-melting"
    )
    dbdf = populate_db(
        dbdf, settings.secondary_structure_paths, parse_secondary, "OligoArrayAux"
    )

    dbdf = reduce_sequence_columns(dbdf)
    dbdf, dtype_sequence = parse_sequences(dbdf)
    dbdf, dtype_header = parse_record_headers(dbdf, prefix)

    dtype = {}
    dtype.update(dtype_melting)
    dtype.update(dtype_hush)
    dtype.update(dtype_secondary)
    dtype.update(dtype_sequence)
    dtype.update(dtype_header)

    for column in dtype:
        if column not in dbdf.columns:
            dbdf["column"] = np.repeat(np.nan, dbdf.shape[0])
    dbdf = dbdf.loc[:, database_columns]

    write_database(dbdf, dtype, settings)

    logging.info("Done. :thumbs_up: :smiley:")


def parse_input(
    path_list: Set[str], parse_function: Callable, software_name: str
) -> pd.DataFrame:
    if not path_list:
        return pd.DataFrame()
    logging.info(f"parsing {software_name} output")
    return pd.concat([parse_function(path) for path in path_list])


def populate_db(
    db: pd.DataFrame,
    path_list: Set[str],
    parse_function: Callable,
    software_name: str,
) -> pd.DataFrame:
    if not path_list:
        return db
    parsed_db = parse_input(path_list, parse_function, software_name)
    logging.info(f"adding {software_name} output to database")
    return db.merge(
        parsed_db,
        how="outer",
        left_index=True,
        right_index=True,
    )


def reduce_sequence_columns(df: pd.DataFrame) -> pd.DataFrame:
    logging.info("discarding redundant sequence columns")
    seq_columns = [c for c in df.columns if "sequence" in c]
    if len(seq_columns) == 1:
        return df
    df.drop(seq_columns[1:], axis=1, inplace=True)
    df.rename(columns={seq_columns[0]: "sequence"}, inplace=True)
    return df


def parse_sequences(df: pd.DataFrame) -> Tuple[pd.DataFrame, Dict[str, str]]:
    logging.info("adding sequence feature columns: length and GC-content")
    sequence_length_list: List[int] = []
    gc_content_list: List[float] = []

    for sequence in track(
        df["sequence"].values, description="calculating GC-content", transient=True
    ):
        sequence = sequence.upper()
        sequence_length_list.append(len(sequence))
        gc_content_list.append(
            (sequence.count(b"G") + sequence.count(b"C")) / len(sequence)
        )

    df["gc_content"] = gc_content_list
    dtype = copy.copy(dtype_sequence_features)
    dtype["sequence"] = f"|S{max(sequence_length_list)}"
    return (df.astype(dtype), dtype)


def parse_record_headers(
    db: pd.DataFrame, chromosome_prefix: str = ""
) -> Tuple[pd.DataFrame, Dict[str, str]]:
    logging.info("adding header feature columns: name, chromosome, start, end")
    name_list: List[str] = []
    name_length_set: Set[int] = set()
    chromosome_list: List[str] = []
    chromosome_length_set: Set[int] = set()
    start_list: List[int] = []
    end_list: List[int] = []

    for record in track(
        db.itertuples(),
        total=db.shape[0],
        description="parsing record headers",
        transient=True,
    ):
        name, position = record.Index.split(" ")
        name_list.append(name)
        name_length_set.add(len(name))
        chromosome, extremes = position.split("=")[1].split(":")
        chromosome_list.append(f"{chromosome_prefix}{chromosome}")
        chromosome_length_set.add(len(f"{chromosome_prefix}{chromosome}"))
        start, end = [int(x) for x in extremes.split("-")]
        if (end - start) != len(record.sequence):
            raise AssertionError(f"{end - start} != {len(record.sequence)}")
        start_list.append(start)
        end_list.append(end)

    db["name"] = name_list
    db["chromosome"] = chromosome_list
    db["start"] = start_list
    db["end"] = end_list
    ass.ert_in_dtype(db["start"].values.max(), "u4")
    ass.ert_in_dtype(db["end"].values.max(), "u4")
    db.reset_index(drop=True, inplace=True)

    dtype = copy.copy(dtype_header_features)
    dtype["name"] = f"|S{max(name_length_set)}"
    dtype["chromosome"] = f"|S{max(chromosome_length_set)}"
    return (db.astype(dtype), dtype)


def write_database(
    dbdf: pd.DataFrame, dtype: Dict[str, str], settings: DBMakeSettings
) -> None:
    with Progress() as progress:
        chromosome_set: Set[bytes] = set(dbdf["chromosome"].values)
        chromosome_data = ChromosomeDict(settings.bin_size)
        chromosome_task = progress.add_task(
            "exporting chromosome",
            total=len(chromosome_set),
            transient=True,
        )
        for selected_chrom in chromosome_set:
            chromosome_db = dbdf.loc[selected_chrom == dbdf["chromosome"], :]

            logging.info(f"sorting records for {selected_chrom.decode()}")
            chromosome_db.sort_values(
                by="start", axis=0, kind="mergesort", inplace=True
            )
            logging.info(f"building index for {selected_chrom.decode()}")
            chromosome_data.add_chromosome(chromosome_db, dtype, progress)

            with open(
                os.path.join(settings.output_path, f"{selected_chrom.decode()}.bin"),
                "wb",
            ) as IH:
                writing_track = progress.add_task(
                    f"writing {selected_chrom.decode()}.bin",
                    total=chromosome_db.shape[0],
                    transient=True,
                )
                for record in chromosome_db.to_records(
                    index=False, column_dtypes=dtype
                ):
                    IH.write(record.tobytes())
                    progress.update(writing_track, advance=1)
            progress.update(chromosome_task, advance=1)

    logging.info("writing db.pickle")
    with open(os.path.join(settings.output_path, "db.pickle"), "wb") as OH:
        pickle.dump(
            dict(chromosomes=chromosome_data, dtype=dtype, args=settings.asdict()), OH
        )



"""
@author: Gabriele Girelli
@contact: gigi.ga90@gmail.com
"""

from Bio.Seq import reverse_complement  # type: ignore
import click  # type: ignore
from ..const import CONTEXT_SETTINGS
from ..fasta import extract_kmers
from ..io import write_oligos
from ..oligo import select_by_GC
import logging
from os.path import isdir, isfile, join as path_join
from os.path import basename, normpath, splitext


def main(input_path: str, output_path: str, kmer_size: int, gcfilter: bool) -> None:
    if not isfile(input_path):
        raise AssertionError
    if not isdir(output_path):
        raise AssertionError

    logging.info(f"Input  : {input_path}")
    logging.info(f"Output : {output_path}")
    logging.info(f"Length : {kmer_size}")

    oligos_list = extract_kmers(input_path, kmer_size)
    logging.info(f"Extracted {len(oligos_list)} sequences")

#    if (gcfilter==False):
    if not gcfilter:
        logging.info(f"Skipped filtering on GC-content.")
        valid_oligos = oligos_list
    else:
        valid_oligos = select_by_GC(oligos_list, kmer_size)
        logging.info(f"{len(valid_oligos)} sequences with correct GC content")
            
    base, _ = splitext(basename(input_path))
    write_oligos(
        path_join(normpath(output_path), f"{base}.GC35to85_Reference.fa"),
        iter(valid_oligos),
    )

    write_oligos(
        path_join(normpath(output_path), f"{base}.GC35to85_RevCompl.fa"),
        ((h, reverse_complement(s)) for h, s in valid_oligos),
        "writing rc",
    )

    logging.info("Done. :thumbs_up: :smiley:")
    logging.shutdown()

    return
